###
# This file contains experimental code snippets for implementing the GFisher test in Hail
# Strategy is to follow the SKAT test in Hail for data processing
# and then implement calculations for Zscore, LD matrix, and GFisher test
# Creator: ZWu 2024-04-10
###

import hail as hl
import numpy as np
import pandas as pd

hl.version()

### ==== Get data in matrix table format
# It must contain genotypes, phenotypes, group
# Optimal: covariates (default can be 1), weights (default can be 1).

hl.reset_global_randomness()
mt = hl.balding_nichols_model(1, n_samples=100, n_variants=20)
mt = mt.annotate_rows(gene=mt.locus.position // 12)
mt = mt.annotate_rows(weight=1)
mt = mt.annotate_cols(phenotype=hl.agg.sum(mt.GT.n_alt_alleles()) - 20 + hl.rand_norm(0, 1))

### ==== process data, and organize them by group
group = mt.gene
weight = mt.weight
y = mt.phenotype  # y corresponds to the original phenotype
x = mt.GT.n_alt_alleles()  # x corresp to the 0, 1, 2 genotypes based on the 0/0, 0/1, 1/1 genotypes (original GT field)
covariates = [1.0]
max_size = 46340
accuracy = 1e-6
iterations = 10000

mt = hl.matrix_table_source('skat/x', x)  # source back to the matrix table

k = len(covariates)  # number of covariates need to be at least 1
if k == 0:
    raise ValueError('At least one covariate is required.')
# _warn_if_no_intercept('_linear_skat', covariates) #??NameError: name '_warn_if_no_intercept' is not defined

# redefine the matrix table, set up column fields (y, covariates), row fields (group, weight)
mt = mt._select_all(
    row_exprs=dict(group=group, weight=weight),
    col_exprs=dict(y=y, covariates=covariates),  # covariates extends [1.0] to [1.0, 1.0, 1.0, ...] for all samples
    entry_exprs=dict(x=x)
)

# Retrieve data: yvec, covmat, n_complete_samples.
yvec, covmat, n = mt.aggregate_cols(
    (hl.agg.collect(hl.float(mt.y)), hl.agg.collect(mt.covariates.map(hl.float)), hl.agg.count()),
    _localize=False
)

# Add global fields: yvec and covmat (as ndarrays), and n_complete_samples.
mt = mt.annotate_globals(yvec=hl.nd.array(yvec),
                         covmat=hl.nd.array(covmat),
                         n_complete_samples=n)

# QR decomposition (return expressions of ndarrays), see https://hail.is/docs/0.2/nd/index.html#hail.nd.qr
covmat_Q, R = hl.nd.qr(mt.covmat)

# Add covmat_Q to the global fields.
# Why not directly make the calculation using covmat_Q?
mt = mt.annotate_globals(covmat_Q=covmat_Q)

null_mu = mt.covmat_Q @ (mt.covmat_Q.T @ mt.yvec)  # fitted responses under the null model

y_residual = mt.yvec - null_mu  # residuals
# store residuals and variance of residuals in global fields
mt = mt.annotate_globals(y_residual=y_residual, s2=y_residual @ y_residual.T / (n - k))

##=== Calculate the Q statistics
# Replace missing genotype values in x with the mean of x
mt = mt.annotate_rows(G_row_mean=hl.agg.mean(mt.x))  # G_row_mean is the mean of x for each row = MAF*2
# mt.describe()

# hail.expression.function.coalesce(*arg) returns the first non-missing value of args, see https://hail.is/docs/0.2/functions/core.html#hail.expr.functions.coalesce
# so hl.coalesce(mt.x, mt.G_row_mean) return the mean of x if x is missing
# so hl.agg.collect(hl.coalesce(mt.x, mt.G_row_mean)) collects x values while replacing missing values with the mean
# each row of G_row is an array of genotype of the variant
mt = mt.annotate_rows(G_row=hl.agg.collect(hl.coalesce(mt.x, mt.G_row_mean)))
# mt.x.show() #genotype values of the entry field are still there
# mt.describe()
# mt.G_row.show() #In the row fields, G_row is a list/array of genotype for each row.

# ht is a Table object created from the global and row fields of mt.
# See https://hail.is/docs/0.2/hail.Table.html
# .    https://hail.is/docs/0.2/overview/table.html
ht = mt.rows()
# ht.describe()
# ht.show()

# filter rows with missing group or weight
ht = ht.filter(hl.all(hl.is_defined(ht.group), hl.is_defined(ht.weight)))
# ht.describe()
# ht.show(20)
# ht.count()
# ht.G_row.show(20)

# grouping by 'group', aggregate the weight and G_row and cast them into arrays
# .aggregate() is used to compute an aggregation over a group. See https://hail.is/docs/0.2/hail.GroupedTable.html#hail.GroupedTable.aggregate
# .agg.take(expr, n, ordering=None) is to take n records of expr (cast into an array),
#    optionally ordered by ordering. See https://hail.is/docs/0.2/aggregators.html#hail.expr.aggregators.take
# After this step, ht only has these fields (locus and allels were not there anymore): group, weight_take, G_take, size
# The key is 'group'
# max_size = 10
ht = ht.group_by('group').aggregate(
    weight_take=hl.agg.take(ht.weight, n=max_size + 1),  # weight_take is a list of weights in each group
    G_take=hl.agg.take(ht.G_row, n=max_size + 1),  # for each group, G_take is a list of genotype lists based on G_row
    size=hl.agg.count(),
)
# ht.describe()
# ht.count() #only two groups/rows left
# ht.show(20) #only two rows left corresp to two groups
# ht.G_take.show() #for each group, G_take is a list of genotype lists (the type is said to be 'array<array<float64>>')
# hl.len(ht.G_take).show() #show the length of G_take array; it gives the group sizes
# hl.len(ht.weight_take).show() #show the length of weight_take array; it gives the group sizes


# This step is to 1) avoid calcu too large group, 2) cast the weight and G_take into ndarray format
# .or_missing(predicate, value): Returns value if predicate is True, otherwise returns missing. See https://hail.is/docs/0.2/functions/core.html#hail.expr.functions.or_missing
# I.e., if the length of weight_take is more than max_size, then weight is missing
# .nd.array() is to cast an array into a nd array. See https://hail.is/docs/0.2/nd/index.html#hail.nd.array
# 'weight' and 'G' are ndarray, i.e., n-dimensional numpy array, see https://numpy.org/doc/stable/reference/arrays.ndarray.html
# .ndarray.T is to transpose the array
ht = ht.annotate(
    weight=hl.nd.array(hl.or_missing(hl.len(ht.weight_take) <= max_size, ht.weight_take)),
    G=hl.nd.array(hl.or_missing(hl.len(ht.G_take) <= max_size, ht.G_take)).T,
)
# ht.describe()
# ht.show()
# ht.weight.show()
# hl.len(ht.weight).show()
# ht.G.show() #For each group, G is a matrix, i.e., ndarray with shape being (sample size, group size)


# This is where the calculation of the Q statistics is applied to all groups/genes
# Calculate the Q statistics, broadcasted over groups (since ht.G and ht.weight have ndarray values over groups).
# ht.y_residual is the same over groups (it is a global field). It is 1-dim ndarray, with shape being (sample size,)
# ht.G is the genotype matrix for each group,
# so (ht.y_residual @ ht.G) is a 1-dim ndarray for each group, with shape being (group size)
# (ht.y_residual @ ht.G).map(lambda x: x**2) is to square each element of the ndarray, with shape being (group size)
# ht.weight is 1-dim array of weights for each group, with shape being (group size)
# so (ht.y_residual @ ht.G).map(lambda x: x**2)* ht.weight is for component-wise multiplication of the two ndarrays
# .   for each group, with shape being (group size)
# .sum(0) is to sum the ndarray along the 0-th axis. See https://hail.is/docs/0.2/hail.expr.NDArrayNumericExpression.html#hail.expr.NDArrayNumericExpression.sum
ht = ht.annotate(Q=((ht.y_residual @ ht.G).map(lambda x: x ** 2) * ht.weight).sum(0))
# ht.describe()
# ht.show()
# ht.y_residual.show()
# (ht.y_residual @ ht.G).show() #two ndarrays, one for each group
# (ht.y_residual @ ht.G).map(lambda x: x**2).show() #two ndarrays, one for each group
# ((ht.y_residual @ ht.G).map(lambda x: x**2) * ht.weight).show() #two ndarrays, one for each group
# ((ht.y_residual @ ht.G).map(lambda x: x**2) * ht.weight).sum(0).show() #two sums, i.e., Q stat, one for each group
# ht.Q.show() #two Q stat, one for each group

###==== Calculate Z scores and correlation matrices (same outupt as Func_Calcu_Z_Scores.py)
GHG = ht.G.T @ ht.G - ht.G.T @ ht.covmat_Q @ ht.covmat_Q.T @ ht.G  # GHG is the covariance matrix for G given X (covariates)
GHG.show()
scores = (ht.y_residual @ ht.G) / hl.sqrt(ht.s2)  # The marginal score statistics
scores.show()
(scores.map(lambda x: x ** 2) * ht.weight).sum(0).show()  # This gives the SKAT statistic, same as below
(ht.Q / ht.s2).show()  # This how Hail gives the SKAT statistic, same as above (without devision by 2)

# Convert GHG into correlation matrix
# hl.nd.diagonal() is to get the diagonal elements of a matrix. See https://hail.is/docs/0.2/nd/index.html#hail.nd.diagonal
sd_GHG = hl.sqrt(hl.nd.diagonal(GHG))  # the std devs based on the diagonal elements of GHG
sd_GHG.show()
sd_GHG.dtype

# Z scores based on the score statistics
Zscores = scores / sd_GHG
Zscores.show()

# Hail statistical functions: https://hail.is/docs/0.2/functions/stats.html#statistical-functions
# apply each element of the Zscores ndarray to the pnorm function
pvalues = Zscores.map(lambda x: 2 * hl.pnorm(-hl.abs(x), lower_tail=True))
pvalues.show()
(hl.pnorm(-1.58) * 2).show()  # p-values based on the Z scores
pvalues.dtype

# convert sd_GHG into a matrix of one row
sd_GHG_m = sd_GHG.reshape((1, sd_GHG.shape[0]))
sd_GHG_m.show()
outer_sd_GHG = sd_GHG_m.T @ sd_GHG_m  # outer product of the diagonal elements
outer_sd_GHG.show()
M_s = GHG / outer_sd_GHG
M_s.show()  # M_s is the correlation matrix for G given X (covariates)



#TODO what exactly does this portion of the code in terms of the pipeline?
##==Calculate the GFisher test statistics in general format
# based on a given function g(Zscores), imitating the R function calcu_SgZ_p() in file Func_Calcu_Comb_Tests.R.

##By quadratic function, we get quadratic statistics
# This corresponds to SKAT as a special case with specific weights.
def g_quadr(x):  # quadradic function
    return x ** 2


g = g_quadr
gf_quadr = (Zscores.map(lambda x: g(x)) * ht.weight).sum(0)  # directly use weights in ht
gf_quadr.show()

# We can get SKAT by using Zscores, gquadr, and proper weights being sd_GHG.map(lambda x: x**2)
(scores.map(lambda x: g(x)) * ht.weight).sum(0).show()  # This gives the SKAT statistic, same as above
(Zscores.map(lambda x: g(x)) * sd_GHG.map(lambda x: x ** 2)).sum(
    0).show()  # This gives the SKAT statistic, same as above


##By identity function, we get Liptak test statistics
# This corresponds to the burden test as a special case
def g_ident(x):  # identity function
    return x


g = g_ident
gf_ident = (Zscores.map(lambda x: g(x)) * ht.weight).sum(0)  # directly use weights in ht
gf_ident.show()


##By GFisher transformation function, we get GFisher test statistics
def g_GFisher_two(x, df=2):
    return hl.qchisqtail(hl.log(2) + hl.pnorm(-hl.abs(x), log_p=True), df=df, lower_tail=False, log_p=True)


# Fisher's combination test statistic
g = g_GFisher_two
g(1).show()
gf_GFisher = (Zscores.map(lambda x: g(x)) * ht.weight).sum(0)  # directly use weights in ht
print("gf-GFisher stat is ")
gf_GFisher.show()


# SKAT statistic
def gg(x):
    return g_GFisher_two(x, df=1)


gf_SKAT = (Zscores.map(lambda x: gg(x)) * sd_GHG.map(lambda x: x ** 2)).sum(0)
gf_SKAT.show()  # This gives the SKAT statistic, same as above

###==== end of the code snippet





# Convert ht.weight to an array<int32> type. However, it seems not necessary (see below).
# ht.weight was ndarray<int32, 1> type. weights_arr is array<int32> type
# ht.weight.show()
weights_arr = hl.array(ht.weight)
# weights_arr.show()
# weights_arr.dtype #array<int32>

# If we keep the original type, the following calculation of A will still work!
weights_arr = ht.weight
weights_arr.show()

# The following is to calculate the above described matrix
# A = \sigma (I - Q Q.T) G diag(sqrt(w)) if all weights are positive
# else, return an error message
# hl.case() is to create a conditional expression. See https://hail.is/docs/0.2/functions/core.html#hail.expr.functions.case
# .when() is to create a conditional branch. See https://hail.is/docs/0.2/functions/core.html#hail.expr.functions.when
# .or_error() is to raise an error if the condition is not met. See https://hail.is/docs/0.2/functions/core.html#hail.expr.functions.or_error
# hl.all() is to Check for all True in boolean expressions or collections of booleans. See https://hail.is/docs/0.2/functions/collections.html#hail.expr.functions.all
A = (
    hl.case()
    .when(
        hl.all(weights_arr.map(lambda x: x >= 0)),  # all weights must be positive
        (ht.G - ht.covmat_Q @ (ht.covmat_Q.T @ ht.G)) * hl.sqrt(ht.weight),  # A = \sigma (I - Q Q.T) G diag(sqrt(w))
    )
    .or_error(
        hl.format(
            'hl._linear_skat: every weight must be positive, in group %s, the weights were: %s',
            ht.group,
            weights_arr,
        )
    )
)
A.show()  # A has an ndarray (2-dim matrix) for each group
A.shape.show()

singular_values = hl.nd.svd(A, compute_uv=False)
singular_values.show()  # 1-dim ndarray for each group
singular_values.shape.show()

# SVD(M) = U S V. U and V are unitary, therefore SVD(k M) = U (k S) V.
eigenvalues = ht.s2 * singular_values.map(lambda x: x ** 2)
eigenvalues.show()  # 1-dim ndarray for each group
eigenvalues.shape.show()
hl.min(eigenvalues).show()

# The R implementation of SKAT, Function.R, Get_Lambda_Approx filters the eigenvalues,
# presumably because a good estimate of the Generalized Chi-Sqaured CDF is not significantly
# affected by chi-squared components with very tiny weights.
threshold = 1e-5 * eigenvalues.sum() / eigenvalues.shape[0]
threshold.show()

# w is the array of weights for each non-central chi-square term.
w = hl.array(eigenvalues).filter(lambda y: y >= threshold)
# w = hl.array(eigenvalues).filter(lambda y: y >= 300) #threshold) #filter could reduce the number of weights
w.show()
hl.len(w).show()

# Calcu p-value using func pgenchisq, the left-tail prob of the generalized chi-squared distribution
# See https://hail.is/docs/0.2/functions/stats.html#hail.expr.functions.pgenchisq
# ??? what does _func("pgenchisq"...) mean in the source code: https://hail.is/docs/0.2/_modules/hail/expr/functions.html#pgenchisq
# ??? where is the real source code of pgenchisq?
genchisq_data = hl.pgenchisq(
    ht.Q,
    w=w,
    k=hl.nd.ones(hl.len(w), dtype=hl.tint32),
    lam=hl.nd.zeros(hl.len(w)),
    mu=0,
    sigma=0,
    min_accuracy=accuracy,
    max_iterations=iterations,
)
genchisq_data.show()

ht = ht.select(
    'size',
    # for reasons unknown, the R implementation calls this expression the Q statistic (which is
    # *not* what they write in the paper)
    q_stat=ht.Q / 2 / ht.s2,
    # The reasoning for taking the complement of the CDF value is:
    #
    # 1. Q is a measure of variance and thus positive.
    #
    # 2. We want to know the probability of obtaining a variance even larger ("more extreme")
    #
    # Ergo, we want to check the right-tail of the distribution.
    p_value=1.0 - genchisq_data.value,
    fault=genchisq_data.fault,
)
ht.show()
ht.describe()

# When return the result, keep global fields: y_residual, s2, n_complete_samples only
# all row fields are remained.
ht.select_globals('y_residual', 's2', 'n_complete_samples')