###
# This file contains experimental code snippets for implementing the GFisher test in Hail
# Strategy is to follow the SKAT test in Hail for data processing
# and then implement calculations for Zscore, LD matrix, and GFisher test
# Creator: ZWu 2024-04-10
# Modified: TYakovenko 2024-04-19
###

import hail as hl
import pyspark.pandas as ps


hl.version()
hl.init(log="hailDistributedProcessingLog.log")

#using the example given in the Hail source code for Skat, output the ht matrix table once it has been processed
def saveHt ():
    ### ==== Get data in matrix table format
    # It must contain genotypes, phenotypes, group
    # Optimal: covariates (default can be 1), weights (default can be 1).

    hl.reset_global_randomness()
    mt = hl.balding_nichols_model(1, n_samples=100, n_variants=20)
    mt = mt.annotate_rows(gene=mt.locus.position // 12)
    mt = mt.annotate_rows(weight=1)
    mt = mt.annotate_cols(phenotype=hl.agg.sum(mt.GT.n_alt_alleles()) - 20 + hl.rand_norm(0, 1))

    ### ==== process data, and organize them by group
    group = mt.gene
    weight = mt.weight
    y = mt.phenotype  # y corresponds to the original phenotype
    x = mt.GT.n_alt_alleles()  # x corresp to the 0, 1, 2 genotypes based on the 0/0, 0/1, 1/1 genotypes (original GT field)
    covariates = [1.0]
    max_size = 46340
    accuracy = 1e-6
    iterations = 10000
    mt = hl.matrix_table_source('skat/x', x)  # source back to the matrix table

    k = len(covariates)  # number of covariates need to be at least 1
    if k == 0:
        raise ValueError('At least one covariate is required.')
    # _warn_if_no_intercept('_linear_skat', covariates) #??NameError: name '_warn_if_no_intercept' is not defined

    # redefine the matrix table, set up column fields (y, covariates), row fields (group, weight)
    mt = mt._select_all(
        row_exprs=dict(group=group, weight=weight),
        col_exprs=dict(y=y, covariates=covariates),  # covariates extends [1.0] to [1.0, 1.0, 1.0, ...] for all samples
        entry_exprs=dict(x=x)
    )

    # Retrieve data: yvec, covmat, n_complete_samples.
    yvec, covmat, n = mt.aggregate_cols(
        (hl.agg.collect(hl.float(mt.y)), hl.agg.collect(mt.covariates.map(hl.float)), hl.agg.count()),
        _localize=False
    )

    # Add global fields: yvec and covmat (as ndarrays), and n_complete_samples.
    mt = mt.annotate_globals(yvec=hl.nd.array(yvec),
                             covmat=hl.nd.array(covmat),
                             n_complete_samples=n)

    # QR decomposition (return expressions of ndarrays), see https://hail.is/docs/0.2/nd/index.html#hail.nd.qr
    covmat_Q, R = hl.nd.qr(mt.covmat)

    # Add covmat_Q to the global fields.
    # Why not directly make the calculation using covmat_Q?
    mt = mt.annotate_globals(covmat_Q=covmat_Q)
    null_mu = mt.covmat_Q @ (mt.covmat_Q.T @ mt.yvec)  # fitted responses under the null model
    y_residual = mt.yvec - null_mu  # residuals
    # store residuals and variance of residuals in global fields
    mt = mt.annotate_globals(y_residual=y_residual, s2=y_residual @ y_residual.T / (n - k))

    ##=== Calculate the Q statistics
    # Replace missing genotype values in x with the mean of x
    mt = mt.annotate_rows(G_row_mean=hl.agg.mean(mt.x))  # G_row_mean is the mean of x for each row = MAF*2
    mt = mt.annotate_rows(G_row=hl.agg.collect(hl.coalesce(mt.x, mt.G_row_mean)))
    # ht is a Table object created from the global and row fields of mt.
    # See https://hail.is/docs/0.2/hail.Table.html
    # .    https://hail.is/docs/0.2/overview/table.html
    ht = mt.rows()
    # filter rows with missing group or weight
    ht = ht.filter(hl.all(hl.is_defined(ht.group), hl.is_defined(ht.weight)))

    # grouping by 'group', aggregate the weight and G_row and cast them into arrays
    # .aggregate() is used to compute an aggregation over a group. See https://hail.is/docs/0.2/hail.GroupedTable.html#hail.GroupedTable.aggregate
    # .agg.take(expr, n, ordering=None) is to take n records of expr (cast into an array),
    #    optionally ordered by ordering. See https://hail.is/docs/0.2/aggregators.html#hail.expr.aggregators.take
    # After this step, ht only has these fields (locus and allels were not there anymore): group, weight_take, G_take, size
    # The key is 'group'
    # max_size = 10
    ht = ht.group_by('group').aggregate(
        weight_take=hl.agg.take(ht.weight, n=max_size + 1),  # weight_take is a list of weights in each group
        G_take=hl.agg.take(ht.G_row, n=max_size + 1),
        # for each group, G_take is a list of genotype lists based on G_row
        size=hl.agg.count(),
    )

    # This step is to 1) avoid calcu too large group, 2) cast the weight and G_take into ndarray format
    # .or_missing(predicate, value): Returns value if predicate is True, otherwise returns missing. See https://hail.is/docs/0.2/functions/core.html#hail.expr.functions.or_missing
    # I.e., if the length of weight_take is more than max_size, then weight is missing
    # .nd.array() is to cast an array into a nd array. See https://hail.is/docs/0.2/nd/index.html#hail.nd.array
    # 'weight' and 'G' are ndarray, i.e., n-dimensional numpy array, see https://numpy.org/doc/stable/reference/arrays.ndarray.html
    # .ndarray.T is to transpose the array
    ht = ht.annotate(
        weight=hl.nd.array(hl.or_missing(hl.len(ht.weight_take) <= max_size, ht.weight_take)),
        G=hl.nd.array(hl.or_missing(hl.len(ht.G_take) <= max_size, ht.G_take)).T,
    )

    # This is where the calculation of the Q statistics is applied to all groups/genes
    # Calculate the Q statistics, broadcasted over groups (since ht.G and ht.weight have ndarray values over groups).
    # ht.y_residual is the same over groups (it is a global field). It is 1-dim ndarray, with shape being (sample size,)
    # ht.G is the genotype matrix for each group,
    # so (ht.y_residual @ ht.G) is a 1-dim ndarray for each group, with shape being (group size)
    # (ht.y_residual @ ht.G).map(lambda x: x**2) is to square each element of the ndarray, with shape being (group size)
    # ht.weight is 1-dim array of weights for each group, with shape being (group size)
    # so (ht.y_residual @ ht.G).map(lambda x: x**2)* ht.weight is for component-wise multiplication of the two ndarrays
    # .   for each group, with shape being (group size)
    # .sum(0) is to sum the ndarray along the 0-th axis. See https://hail.is/docs/0.2/hail.expr.NDArrayNumericExpression.html#hail.expr.NDArrayNumericExpression.sum
    ht = ht.annotate(Q=((ht.y_residual @ ht.G).map(lambda x: x ** 2) * ht.weight).sum(0))
    #here the storage of the data is temporary and is only neede to simplify the code calls below as the data processing will all be the same
    ht.write("ht.mt")

ht = hl.read_table("../hailMatrices/ht.mt")

#function to return a hail matrix with all the necessary pGFisher inputs
def findGFM (ht):
    # calculate the Z statistics needed for the pGFisher function call

    ###==== Calculate Z scores and correlation matrices (same outupt as Func_Calcu_Z_Scores.py)
    GHG = ht.G.T @ ht.G - ht.G.T @ ht.covmat_Q @ ht.covmat_Q.T @ ht.G  # GHG is the covariance matrix for G given X (covariates)
    scores = (ht.y_residual @ ht.G) / hl.sqrt(ht.s2)  # The marginal score statistics
    # (scores.map(lambda x: x ** 2) * ht.weight).sum(0).show()  # This gives the SKAT statistic, same as below
    # (ht.Q / ht.s2).show()  # This how Hail gives the SKAT statistic, same as above (without devision by 2)
    # Convert GHG into correlation matrix
    # hl.nd.diagonal() is to get the diagonal elements of a matrix. See https://hail.is/docs/0.2/nd/index.html#hail.nd.diagonal
    sd_GHG = hl.sqrt(hl.nd.diagonal(GHG))  # the std devs based on the diagonal elements of GHG
    # Z scores based on the score statistics
    Zscores = scores / sd_GHG
    '''
    the calculations here and above are Hail based and follow Skat fairly closely
    '''
    #print("The Z statistic is as follows:")
    #Zscores.show()

    # Hail statistical functions: https://hail.is/docs/0.2/functions/stats.html#statistical-functions
    # apply each element of the Zscores ndarray to the pnorm function
    pvalues = Zscores.map(lambda x: 2 * hl.pnorm(-hl.abs(x), lower_tail=True))
    # TODO: what is the difference between the pvalues on line 143 and 144 for our implementation/calls?
    #The above p values are p values for each SNP;
    #for gfisher stat we can calculate it using pvalue or Zscore (the calculation is the same)
    #in the R code, we allow for both/either inputs, here we now have the flexibility to do the pvalue input or Zscore input

    pvalues.show()  # these pvalues are the onse that will go into the stat_GFisher which is the only thing needed to call the other p value calculation
    #(hl.pnorm(-1.58) * 2).show()  # p-values based on the Z scores
    '''
    hail_Trial_Wu contains more calculations necessary for comparison of various tests
    the functions below are more general gFisher statistics
    '''
    # convert sd_GHG into a matrix of one row
    #TODO double check if the reshaping here works with distributed computing
    #the reshaping below is done st we are working with a matrix instrad of a vector
    sd_GHG_m = sd_GHG.reshape(
        (1, sd_GHG.shape[0]))  # method reshape doesn't allign with the calss Float here(?????????????????????)
    # sd_GHG_m.show()
    outer_sd_GHG = sd_GHG_m.T @ sd_GHG_m  # outer product of the diagonal elements
    #^requires matrices and reshaping
    # outer_sd_GHG.show()
    M_s = GHG / outer_sd_GHG
    #print("The M matrix is")
    #M_s.show()  # M_s is the correlation matrix for G given X (covariates)
    #annotate the global fields in the input data -> there is NO return here; we are adding onto a matrix in place
    #no annotation in place but we can replace it such that there are no returns
    # Annotate multiple global fields to the MatrixTable
    def g_quadr(x):  # quadradic function
        return x ** 2

    g = g_quadr
    gf_quadr = (Zscores.map(lambda x: g(x)) * ht.weight).sum(0)  # directly use weights in ht
    #gf_quadr.show()

    # We can get SKAT by using Zscores, gquadr, and proper weights being sd_GHG.map(lambda x: x**2)
    (scores.map(lambda x: g(x)) * ht.weight).sum(0)  # This gives the SKAT statistic, same as above
    #(Zscores.map(lambda x: g(x)) * sd_GHG.map(lambda x: x ** 2)).sum(0).show()  # This gives the SKAT statistic, same as above

    ##By identity function, we get Liptak test statistics
    # This corresponds to the burden test as a special case
    def g_ident(x):  # identity function
        return x

    g = g_ident
    gf_ident = (Zscores.map(lambda x: g(x)) * ht.weight).sum(0)  # directly use weights in ht
    #gf_ident.show()

    ##By GFisher transformation function, we get GFisher test statistics
    def g_GFisher_two(x, df=2):
        return hl.qchisqtail(hl.log(2) + hl.pnorm(-hl.abs(x), log_p=True), df=df, lower_tail=False, log_p=True)

    # Fisher's combination test statistic
    g = g_GFisher_two
    gf_GFisher = (Zscores.map(lambda x: g(x)) * ht.weight).sum(0)  # directly use weights in ht
    #TODO we need to slightly modify these calculations such that they can be saved along the way into a matrix table
    '''
    as of right now, we can not annotate a matrix table due to the  fact that all the calculations performed for gf_GFisher and M_s changed the datatype
    '''
    return gf_GFisher, M_s

#TODO: we originally implemented stat_GFisher to calculate the gfisher statistic in Python, do we need to reimplement that here?
#just use one of the statistics above as the pval calculation
gf, M = findGFM(ht) #M_s is directly called in p_GFisher but pvals need to be recalibrated
ls = gf * gf
ls.show()
lf = M @ M #matrix multiplication support
lf.show()
#now figure out how to work with the exact data types in spark if needed
'''
in order to convert this structure, which is an NDArrayNumericExpression in Hail
we need to use evaluate function or some sort of conversion into pandas/numpy/spark
all of which require for all the data to be collected on a single driver node. 
In this manner, we will not gain computational advantages so we need to keep working directly in Hail. 
Another approach is to go further into spark and see if it is possible to construct the needed data type for RDD


The main challenge right now is figuring out how to implement the more complex statistics operations needed for our functions
'''

#TODO Hail has a get_gene_intervals built in function to calculate the gene intervals from a GTF file for quixk filtering -> it might be useful for our grouping
#here are almost all the functions I might need for Hail calculations to be performed
#https://hail.is/docs/0.2/functions/stats.html#hail.expr.functions.chi_squared_test
#cholesky decomposition in spark https://spark.apache.org/docs/2.1.2/api/java/org/apache/spark/mllib/linalg/CholeskyDecomposition.html
#cholesky algorithm description for distributive processing https://ieeexplore.ieee.org/abstract/document/7798939
#NPD class: https://stackoverflow.com/questions/619335/a-simple-algorithm-for-generating-positive-semidefinite-matrices
#ndp further
#hail nd exmaples https://hail.is/docs/0.2/nd/index.html#hail.nd.solve
#instead of Cholesky use some other decomposition



'''
at this point, we want to create a simple example of a function that we use and show that it can interface with Hail to assure distributed computation
^this is a demonstration of feasibility and efficiency
basically show the easiest way to interface with Hail
as an example function implementation, pchisqtail from Hail implementation or pnorm or any that use _func
^see if I can trace their function implementation that they have in Scala
'''