#####################################
# This file contains main and supportive functions for GFisher test, including computing
# the GFisher statistics and p-values.
# Reference: Zhang, Hong, and Zheyang Wu. "The generalized Fisher's combination and accurate p‐value
#            calculation under dependence." Biometrics 79.2 (2023): 1159-1172.
# Creator: Hong Zhang
# Last update:
#   ZWu 2024-02-29: Organize functions by their usages; add some headers and comments;
#                   pval.oGFisher: Add individual GFisher statistics into the output list
#                   p.GFisher: Fix a bug: if(class(M_chol)=="try-error") -> if("try-error" %in% class(M_chol))
#####################################


library(mvtnorm)
library(Matrix)


#####################################
# Main functions for users
#####################################

#' Compute the GFisher test statistics S = sum_i w_i F^{-1}_{d_i}(1-P_i), based on a vector of p-values
#' P_i's, degrees of freedom d_i's, and weights w_i's. F^{-1}_{d_i} is the inverse CDF of the chi-square
#' distribution with d_i degrees of freedom.
#' @author Hong Zhang
#' @param p a vector of input p-values of the GFisher test
#' @param df degrees of freedom for inverse chi-square transformation for each p-value
#'           It can be a vector of the same length as p, indicating each transformation function
#'           might have a different df, or a single number, indicating the same degrees of freedom for all.
#' @param w vector of non-negative weights for each p-value. Default is 1 (equal weights).
#' @return fisherstat: the GFisher test statistics
#' @detail The statistic is calculated based on normalized weights, i.e., w/mean(w).
#' @example
#' n = 10
#' pval = runif(n)
#' stat.GFisher(pval, df=2, w=1)
#' stat.GFisher(pval, df=rep(2,n), w=rep(1,n))
#' stat.GFisher(pval, df=1:n, w=1:n)
#' @export
#'
stat.GFisher = function(p, df=2, w=1){
  if(length(w)>1){
    w = length(p)*w/sum(w) #require non-negative weights o.w. sum(w) may not be well defined; it could be 0
  }else{
    w = 1
  }
  if(all(df==2)){
    pp_trans = -2*log(p)
  }else{
    #pp_trans = qchisq(1-p, df=df)
    pp_trans = qchisq(p, df=df, lower.tail=F)
    if(anyNA(pp_trans)){
      na_id = which(is.na(pp_trans))
      pp_trans[na_id] = qchisq(1-p[na_id], df=df[na_id])}
  }
  fisherstat = sum(w*pp_trans) # apply(pp_trans%*%W, 1, sum)
  return(fisherstat)
}


#' Compute the p-value of a GFisher test
#' @author Hong Zhang
#' @param q  observed GFisher statistic.
#' @param df  vector of degrees of freedom for inverse chi-square transformation for each
#'             p-value. If all df’s are equal, it can be defined by the constant.
#' @param w  vector of non-negative weights.
#' @param M  correlation matrix of the Zscores from which the input p-values were obtained.
#' @param p.type  "two" = two-sided input p-values, "one" = one-sided input p-values.
#' @param method  "MR" = simulation-assisted moment ratio matching,
#'                "HYB" = moment ratio matching by quadratic approximation,
#'                "GB" = Brown’s method with calculated variance. See details in the reference.
#' @param nsim  number of simulation used in the "MR" method, default = 5e4.
#' @param seed  seed for random number generation, default = NULL.
#' @return pval: the p-value of the GFisher test
#' @example
#' set.seed(123)
#'     n = 10
#'     M = matrix(0.3, n, n) + diag(0.7, n, n)
#'     zscore = matrix(rnorm(n),nrow=1)%*%chol(M)
#'     pval = 2*(1-pnorm(abs(zscore))) #two-sided p-values
#'     gf1 = stat.GFisher(pval, df=2, w=1)
#'     gf2 = stat.GFisher(pval, df=1:n, w=1:n)
#'     p.GFisher(gf1, df=2, w=1, M=M, method="HYB")
#'     p.GFisher(gf1, df=2, w=1, M=M, method="MR", nsim=5e4)
#'     p.GFisher(gf2, df=1:n, w=1:n, M=M, method="HYB")
#'     p.GFisher(gf2, df=1:n, w=1:n, M=M, method="MR", nsim=5e4)
p.GFisher = function(q, df, w, M, p.type="two", method="HYB", nsim=NULL, seed=NULL){
  n = dim(M)[1]
  if(length(df)==1){df = rep(df, n)}
  if(length(w)==1){w = rep(w, n)}
  w = n*w/sum(w)
  if(method=="MR"){
    M_chol = try(chol(M))
    #if(class(M_chol)=="try-error"){ #Wu:class(M_chol) could have two elements, "matrix" "array" , so the if condition could complain. Use %in% to fix it.
    if("try-error" %in% class(M_chol)){
      M = as.matrix(nearPD(M, corr=TRUE)$mat)
      M_chol = chol(M)
    }
    if(class(nsim)!="numeric"){
      nsim = 5e4
    }
    if(class(seed)=="numeric"){
      set.seed(seed)
    }
    znull = matrix(rnorm(n*nsim), ncol=n, nrow=nsim)%*%M_chol
    if(p.type=="two"){
      pnull = 2*pnorm(-abs(znull),0,1)
    }else{
      pnull = pnorm(znull,0,1, lower.tail = FALSE)
    }
    if(sd(df)==0){
      if(all(df==2)){
        pp_trans = -2*log(pnull)
      }else{
        pp_trans = qchisq(1-pnull, df=df[1])
      }
    }else{
      pp_trans = t(apply(pnull, 1, function(x)qchisq(1-x, df=df)))
    }
    fishernull = apply(pp_trans, 1, function(x)sum(x*w))
    MM = sapply(1:4, function(x)mean(fishernull^x))
    mu = MM[1]; sigma2 = (MM[2]-MM[1]^2)*nsim/(nsim-1)
    cmu3 = MM[3]-3*MM[2]*MM[1]+2*MM[1]^3
    cmu4 = MM[4]-4*MM[3]*MM[1]+6*MM[2]*MM[1]^2-3*MM[1]^4
    gm = cmu3/sigma2^(3/2)
    kp = cmu4/sigma2^2
    a = 9*gm^2/(kp-3)^2
    pval = pgamma((q-mu)/sqrt(sigma2)*sqrt(a)+a, shape=a,  scale = 1, lower.tail = FALSE)
  }else{
    GM = getGFisherGM(df, w, M, p.type)
    mu = sum(w*df); sigma2 = sum(GM)
    if(method=="HYB"){
      out_GM = getGFisherlam(df, w, M, GM, p.type)
      lam = out_GM$lam;
      c2 = sum(lam^2); c3 = sum(lam^3); c4 = sum(lam^4)
      gm = sqrt(8)*c3/(c2)^(3/2)
      kp = 12*c4/c2^2 + 3
      a = 9*gm^2/(kp-3)^2
    }else{ #Brown's method
      a = mu^2/sigma2
    }
    pval = pgamma((q-mu)/sqrt(sigma2)*sqrt(a)+a, shape=a,  scale = 1, lower.tail = FALSE)
  }
  return(pval)
}

#' Compute the oGFisher related statistics
#' @author Hong Zhang
#' @param p a vector of input p-values of the oGFisher test
#' @param DF matrix of degrees of freedom for inverse chi-square transformation for each
#'         p-value. Each row represents a GFisher test. It allows a matrix of one column, indicating
#'        the same degrees of freedom for all p-values's chi-square transformations.
#' @param W matrix of non-negative weights. Each row represents a GFisher test.
#' @param M correlation matrix of the input Zscores from which the input p-values were obtained.
#' @param p.type "two" = two-sided, "one" = one-sided input p-values.
#' @param method "MR" = simulation-assisted moment ratio matching,
#'                "HYB" = moment ratio matching by quadratic approximation,
#'                 "GB" = Brown’s method with calculated variance.
#' @param nsim  number of simulation used in the "MR" method, default = 5e4
#' @param seed  seed for random number generation, default = NULL
#' @return STAT: the involved GFisher test statistics
#' @return PVAL: the individual p-values of the involved GFisher tests
#' @return minp: the minimum p-value of PVAL
#' @return cct: the Cauchy combination statistics of PVAL
#' @example
#' set.seed(123)
#' n = 10
#' M = matrix(0.3, n, n) + diag(0.7, n, n)
#' zscore = matrix(rnorm(n),nrow=1)%*%chol(M)
#' pval = 2*(1-pnorm(abs(zscore)))
#' DF = rbind(rep(1,n),rep(2,n))
#' W = rbind(rep(1,n), 1:10)
#' stat.oGFisher(pval, DF, W, M, p.type="two", method="HYB", nsim=5e4)
#' stat.oGFisher(pval, DF=rbind(1,2), W, M, p.type="two", method="HYB", nsim=5e4)
#' @export
#'
stat.oGFisher = function(p, DF, W, M, p.type="two", method="HYB", nsim=NULL, seed=NULL){
  STAT = sapply(1:dim(DF)[1],function(x) stat.GFisher(p = p, df=DF[x,], w=W[x,]))
  PVAL = sapply(1:length(STAT),function(x)p.GFisher(STAT[x], df=DF[x,], w=W[x,], M, p.type, method, nsim, seed))
  minp = min(PVAL)
  # if any PVAL > 0.9, change it to 0.9
  PVAL[PVAL>0.9] = 0.9
  # check if there are very small non-zero p values and calcuate the cauchy statistics
  is.small<-(PVAL<1e-15)
  CCTSTAT = PVAL
  if(sum(is.small)==0){cct = mean(tan(pi*(0.5-CCTSTAT)))}
  if(sum(is.small)>0){
  CCTSTAT[!is.small]<-tan((0.5-CCTSTAT[!is.small])*pi)
  CCTSTAT[is.small]<-1/CCTSTAT[is.small]/pi
  cct = mean(CCTSTAT)}
  #cct = mean(tan(pi*(0.5-PVAL)))
  return(list(STAT=STAT, PVAL=PVAL, minp=minp, cct=cct))
}


#' Compute the oGFisher test based on a vector of p-values
#' @author Hong Zhang
#'         Update: ZWu (2024-02-29): Add individual GFisher statistics into the output list
#' @param p a vector of input p-values of the oGFisher test
#' @param DF matrix of degrees of freedom for inverse chi-square transformation for each
#'          p-value. Each row represents a GFisher test. It allows a matrix of one column, indicating
#'         the same degrees of freedom for all p-values's chi-square transformations.
#' @param W matrix of weights. Each row represents a GFisher test.
#' @param M correlation matrix of the input Zscores from which the input p-values were obtained.
#' @param p.type "two" = two-sided, "one" = one-sided input p-values.
#' @param method  "MR" = simulation-assisted moment ratio matching,
#'                "HYB" = moment ratio matching by quadratic approximation,
#'                 "GB" = Brown’s method with calculated variance.
#' @param combine  "cct" = oGFisher using the Cauchy combination method,
#'                 "mvn" = oGFisher using multivariate normal distribution.
#' @param nsim  number of simulation used in the "MR" method, default = 5e4
#' @param seed  seed for random number generation, default = NULL
#' @return pval: the p-value of the oGFisher test
#' @return pval_indi: the individual p-values of the involved GFisher tests
#' @return stat_indi: the involved GFisher statistics
#' @example
#'    set.seed(123)
#'    n = 10
#'    M = matrix(0.3, n, n) + diag(0.7, n, n)
#'    zscore = matrix(rnorm(n),nrow=1)%*%chol(M)
#'    pval = 2*(1-pnorm(abs(zscore)))
#'    DF = rbind(rep(1,n),rep(2,n))
#'    W = rbind(rep(1,n), 1:10)
#'    pval.oGFisher(pval, DF, W, M, p.type="two", method="HYB", combine="cct")
#'    pval.oGFisher(pval, DF=rbind(1,2), W, M, p.type="two", method="HYB", combine="cct")
#' @export
#' @import mvtnorm
pval.oGFisher = function(p, DF, W, M, p.type="two", method="HYB", combine="cct", nsim=NULL, seed=NULL){
  out = stat.oGFisher(p, DF, W, M, p.type="two", method, nsim=NULL, seed=NULL)
  if(combine=="cct"){
    thr = out$cct
    pval = pcauchy(thr,lower.tail = F)
    #pval = 0.5 - atan(thr)/pi
    #pval = ifelse(thr>1e+15, (1/thr)/pi, 0.5 - atan(thr)/pi)
  }else{
    thr = out$minp
    nd = dim(DF)[1]
    COR_GFisher = getGFisherCOR(DD=DF, W=W, M=M, p.type=p.type)
    pval = 1 - pmvnorm(lower=rep(-Inf,nd), upper=qnorm(1-thr), mean=rep(0,nd), corr=COR_GFisher)[1]
  }
  return(list(pval=pval, pval_indi=out$PVAL, stat_indi=out$STAT))
}


#####################################
# Functions for computing algorithms
#####################################

getGFishercoef = function(D, M, p.type="two"){
#Calculate coefficients for the GFisher method
#'
#' This function computes coefficgit ients for the GFisher method based on a given design matrix and contrasts.
#'
#' @param D A vector specifying the contrasts to be computed.
#' @param M The design matrix.
#' @param p.type Type of p-values to use. Default is "two".
#' @return A list of coefficients for the GFisher method.
#' @examples
#' # Example 1: Using default parameters
#' D <- c(1, 2, 3)
#' M <- matrix(rnorm(9), ncol = 3)
#' getGFishercoef(D, M)
#'
#' # Example 2: Specifying p.type
#' getGFishercoef(D, M, p.type = "one")
#'
#' # Example 3: Using a different design matrix
#' M <- matrix(rnorm(12), ncol = 4)
#' getGFishercoef(D, M)
#'
  n = dim(M)[1]
  if(p.type=="two"){
    coeff2 = sapply(D, function(y)integrate(function(x)qchisq(pchisq(x^2,df=1),df=y)*dnorm(x)*(x^2-1), -8,8)$value)
    coeff4 = sapply(D, function(y)integrate(function(x)qchisq(pchisq(x^2,df=1),df=y)*dnorm(x)*(x^4-6*x^2+3), -8,8)$value)
    coeff6 = sapply(D, function(y)integrate(function(x)qchisq(pchisq(x^2,df=1),df=y)*dnorm(x)*(x^6-15*x^4+45*x^2-15), -8,8)$value)
    coeff8 = sapply(D, function(y)integrate(function(x)qchisq(pchisq(x^2,df=1),df=y)*dnorm(x)*(x^8-28*x^6+210*x^4-420*x^2+105), -8,8)$value)
    if(length(D)==1){
      coeff2 = rep(coeff2, n)
      coeff4 = rep(coeff4, n)
      coeff6 = rep(coeff6, n)
      coeff8 = rep(coeff8, n)
    }
    return(list(coeff2=coeff2,coeff4=coeff4,coeff6=coeff6,coeff8=coeff8))
  }else{
    coeff1 = sapply(D, function(y)integrate(function(x)qchisq(pnorm(x),df=y)*dnorm(x)*(x), -8,8)$value)
    coeff2 = sapply(D, function(y)integrate(function(x)qchisq(pnorm(x),df=y)*dnorm(x)*(x^2-1), -8,8)$value)
    coeff3 = sapply(D, function(y)integrate(function(x)qchisq(pnorm(x),df=y)*dnorm(x)*(x^3-3*x), -8,8)$value)
    coeff4 = sapply(D, function(y)integrate(function(x)qchisq(pnorm(x),df=y)*dnorm(x)*(x^4-6*x^2+3), -8,8)$value)
    if(length(D)==1){
      coeff1 = rep(coeff1, n)
      coeff2 = rep(coeff2, n)
      coeff3 = rep(coeff3, n)
      coeff4 = rep(coeff4, n)
    }
    return(list(coeff1=coeff1,coeff2=coeff2,coeff3=coeff3,coeff4=coeff4))
  }
}

getGFishercov = function(D1, D2,W1,W2, M, p.type="two", var.correct = TRUE){
  n = dim(M)[1]
  res1.coeff = getGFishercoef(D1, M, p.type=p.type)
  res2.coeff = getGFishercoef(D2, M, p.type=p.type)
  if(p.type=="two"){
    GM_cross = (M^2/2*res1.coeff$coeff2%*%t(res2.coeff$coeff2)+
                  M^4/24*res1.coeff$coeff4%*%t(res2.coeff$coeff4)+
                  M^6/720*res1.coeff$coeff6%*%t(res2.coeff$coeff6)+
                  M^8/40320*res1.coeff$coeff8%*%t(res2.coeff$coeff8))
    if(var.correct==TRUE){
      dd = diag(M)
      v1 = (res1.coeff$coeff2)^2/2+
        (res1.coeff$coeff4)^2/24+
        (res1.coeff$coeff6)^2/720+
        (res1.coeff$coeff8)^2/40320
      v2 = (res2.coeff$coeff2)^2/2+
        (res2.coeff$coeff4)^2/24+
        (res2.coeff$coeff6)^2/720+
        (res2.coeff$coeff8)^2/40320
      GM_cross = diag(sqrt(2*D1/v1),n,n)%*%GM_cross%*%diag(sqrt(2*D2/v2),n,n)
    }
  }else{
    GM_cross = (M*res1.coeff$coeff1%*%t(res2.coeff$coeff1)+
                  M^2/2*res1.coeff$coeff2%*%t(res2.coeff$coeff2)+
                  M^3/6*res1.coeff$coeff3%*%t(res2.coeff$coeff3)+
                  M^4/24*res1.coeff$coeff4%*%t(res2.coeff$coeff4))
    if(var.correct==TRUE){
      dd = diag(M)
      v1 = (res1.coeff$coeff1)^2+
        (res1.coeff$coeff2)^2/2+
        (res1.coeff$coeff3)^2/6+
        (res1.coeff$coeff4)^2/24
      v2 = (res2.coeff$coeff1)^2+
        (res2.coeff$coeff2)^2/2+
        (res2.coeff$coeff3)^2/6+
        (res2.coeff$coeff4)^2/24
      GM_cross = diag(sqrt(2*D1/v1),n,n)%*%GM_cross%*%diag(sqrt(2*D2/v2),n,n)
    }
  }
  return(sum(GM_cross*(W1%*%t(W2))))
}

getGFisherCOR = function(DD, W, M, var.correct=TRUE, p.type="two"){ # DD is a matrix, row represents a GFisher df
#' Calculate covariance for the GFisher method
#'
#' This function computes covariance for the GFisher method based on a given design matrix and contrasts.
#'
#' @param D1 A vector specifying the contrasts for the first group.
#' @param D2 A vector specifying the contrasts for the second group.
#' @param W1 Vector of weights for the first group.
#' @param W2 Vector of weights for the second group.
#' @param M The design matrix.
#' @param p.type Type of p-values to use. Default is "two".
#' @param var.correct Logical indicating whether to correct for the variance. Default is TRUE.
#' @return Covariance for the GFisher method.
#' @examples
#' # Example 1: Using default parameters
#' D1 <- c(1, 2, 3)
#' D2 <- c(2, 3, 4)
#' W1 <- c(1, 1, 1)
#' W2 <- c(1, 1, 1)
#' M <- matrix(rnorm(9), ncol = 3)
#' getGFishercov(D1, D2, W1, W2, M)
#'
#' # Example 2: Specifying p.type and var.correct
#' getGFishercov(D1, D2, W1, W2, M, p.type = "one", var.correct = FALSE)
#'
#' # Example 3: Using a different design matrix
#' M <- matrix(rnorm(12), ncol = 4)
#' getGFishercov(D1, D2, W1, W2, M)
#'
#' @export
  m = dim(DD)[1] # m is the numer of GFisher
  COV = matrix(NA, ncol=m, nrow=m)
  for(i in 1:m){
    for(j in i:m){
      COV[i,j] = getGFishercov(DD[i,], DD[j,],W[i,],W[j,], M, var.correct=var.correct,
                               p.type=p.type)
    }
  }
  COV[lower.tri(COV)] = t(COV)[lower.tri(COV)]
  return(cov2cor(COV))
}

getGFisherGM = function(D, w, M, p.type="two"){
#' Calculate Grand Mean for the GFisher method
#'
#' This function computes the Grand Mean for the GFisher method based on a given design matrix and contrasts.
#'
#' @param D A vector specifying the contrasts.
#' @param w Vector of weights for each contrast.
#' @param M The design matrix.
#' @param p.type Type of p-values to use. Default is "two".
#' @return Grand Mean for the GFisher method.
#' @examples
#' # Example 1: Using default parameters
#' D <- c(1, 2, 3)
#' w <- c(1, 1, 1)
#' M <- matrix(rnorm(9), ncol = 3)
#' getGFisherGM(D, w, M)
#'
#' # Example 2: Specifying p.type
#' getGFisherGM(D, w, M, p.type = "one")
#'
#' # Example 3: Using a different design matrix
#' M <- matrix(rnorm(12), ncol = 4)
#' getGFisherGM(D, w, M)
#'
#' @export
  n = dim(M)[1]
  if(p.type=="two"){
    coeff2 = sapply(D, function(y)integrate(function(x)qchisq(pchisq(x^2,df=1),df=y)*dnorm(x)*(x^2-1), -8,8)$value)
    coeff4 = sapply(D, function(y)integrate(function(x)qchisq(pchisq(x^2,df=1),df=y)*dnorm(x)*(x^4-6*x^2+3), -8,8)$value)
    coeff6 = sapply(D, function(y)integrate(function(x)qchisq(pchisq(x^2,df=1),df=y)*dnorm(x)*(x^6-15*x^4+45*x^2-15), -8,8)$value)
    coeff8 = sapply(D, function(y)integrate(function(x)qchisq(pchisq(x^2,df=1),df=y)*dnorm(x)*(x^8-28*x^6+210*x^4-420*x^2+105), -8,8)$value)
    GM = M^2/2*coeff2%*%t(coeff2)+M^4/24*coeff4%*%t(coeff4) +
      M^6/720*coeff6%*%t(coeff6)+M^8/40320*coeff8%*%t(coeff8)
    GM = diag(sqrt(2*D)*w,n,n)%*%cov2cor(GM)%*%diag(sqrt(2*D)*w,n,n)
  }else{
    coeff3 = sapply(D, function(y)integrate(function(x)qchisq(pnorm(x),df=y)*dnorm(x)*(x^3-3*x), -8,8)$value)
    coeff4 = sapply(D, function(y)integrate(function(x)qchisq(pnorm(x),df=y)*dnorm(x)*(x^4-6*x^2+3), -8,8)$value)
    coeff2 = sapply(D, function(y)integrate(function(x)qchisq(pnorm(x),df=y)*dnorm(x)*(x^2-1), -8,8)$value)
    coeff1 = sapply(D, function(y)integrate(function(x)qchisq(pnorm(x),df=y)*dnorm(x)*(x), -8,8)$value)
    GM = M*coeff1%*%t(coeff1)+M^2/2*coeff2%*%t(coeff2)+
      M^3/6*coeff3%*%t(coeff3)+M^4/24*coeff4%*%t(coeff4)
    GM = diag(sqrt(2*D)*w,n,n)%*%cov2cor(GM)%*%diag(sqrt(2*D)*w,n,n)
    #GM[GM<0]=0
  }
  return(GM)
}

##[??Wu: Why p.type and tbval were not used at all? ]
getGFisherlam = function(D, w, M, GM, p.type="two"){
#' Calculate eigenvalues for the GFisher method
#'
#' This function computes eigenvalues for the GFisher method based on a given design matrix and contrasts.
#'
#' @param D Degrees of freedom.
#' @param w Vector of weights for each contrast.
#' @param M The design matrix.
#' @param GM The Grand Mean matrix.
#' @param p.type Type of p-values to use. Default is "two".
#' @return A list of eigenvalues for the GFisher method.
#' @examples
#' # Example 1
#' D <- c(1, 2, 3)
#' w <- c(1, 1, 1)
#' M <- matrix(rnorm(9), ncol = 3)
#' GM <- matrix(rnorm(9), ncol = 3)
#' getGFisherlam(D, w, M, GM)
#'
#' # Example 2
#' getGFisherlam(D, w, M, GM, p.type = "one")
#'
#' # Example 3
#' M <- matrix(rnorm(12), ncol = 4)
#' GM <- matrix(rnorm(12), ncol = 4)
#' getGFisherlam(D, w, M, GM)
#'
#' @export
  n = dim(M)[1]
  DM = pmin(matrix(rep(D,n),nrow=n,byrow = FALSE), matrix(rep(D,n),nrow=n,byrow = TRUE))
  M_tilde = sqrt(abs(GM)/DM/2)*sign(M);
  if(any(M_tilde>1)){
    M_tilde[M_tilde>1] = 0.999
  }
  if(any(eigen(M_tilde)$values<1e-10)){
    M_tilde = as.matrix(nearPD(M_tilde, corr = TRUE)$mat)
  }
  M_tilde_chol = chol(M_tilde)

  WM_chol = M_tilde_chol%*%diag(sqrt(w))#M_chol#
  tb = table(D)
  tbval = as.numeric(names(tb))
  lam =eigen(t(WM_chol)%*%WM_chol, symmetric = TRUE, only.values = TRUE)$values
  if(max(D)>1){
    for(i in 2:max(D)){
      Ai = diag(n)
      Did = which(D<i)
      if(length(Did)>0){
        diag(Ai)[Did] = 0
      }
      lam = c(lam,eigen((WM_chol)%*%Ai%*%t(WM_chol), symmetric = TRUE, only.values = TRUE)$values)
      #print(diag(diag(sqrt(w))%*%(Ai)%*%diag(sqrt(w))))
    }
  }
  lam = lam[lam>1e-10]
  return(list(lam=lam))
}