######
# This file contains functions for processing genotype and phenotype data
# Author: Ming Liu, ZWu
#####

###======== Dependents ======== ###
library(caret);



###======== Functions ======== ###

#' Filter Genotype Matrix Based on MAC and LD
#'
#' For a given genotype matrix, this function removes SNP columns with small minor allele count (MAC) or large correlations (LD).
#' It depends on the caret package.
#'
#' @param G a matrix of genotye, the columns are the SNPs, the rows are independent observations.
#' @param mac.thresh a numeric threshold of MAC.
#' @param ld.thresh a numeric threshold of LD.
#' @return A list containing the matrix of genotype after removing SNPs that have too lower MAC or too high LD, 
#'         vector of MAC, minor allele frequency (MAF), and correlation matrix.
#' @examples
#' geno_matrix = matrix(runif(1000), 100, 10)
#' result = filter_genotype_mac_ld(geno_matrix, mac.thresh = 10, ld.thresh = 0.2)
#' result = filter_genotype_mac_ld(geno_matrix, mac.thresh = 50, ld.thresh = 0.9)
#' @import caret
#' @export
filter_genotype_mac_ld <- function(G, mac.thresh=20, ld.thresh=0.9) {
  # Remove SNPs with too low MAC
  MAC = apply(G,2,sum)
  snpNeed = which(MAC>=mac.thresh & !is.na(MAC))
  G = G[,snpNeed,drop=F]
  
  # Remove highly correlated SNPs
  if (dim(G)[2] > 1) {
    corM = cor(G)
    hc = sort(caret::findCorrelation(corM, cutoff = ld.thresh))
    if (length(hc) > 0) {
      G = as.matrix(G[, -hc, drop = F])
    }
  }
  
  #Summary after filtering
  MAC = apply(G, 2, sum) 
  MAF = apply(G, 2, mean) / 2
  corM = cor(G)
  
  return(list(G = G, MAC = as.vector(MAC), MAF = as.vector(MAF), corM = as.matrix(corM)))
}



#' Centralize the genotype data, which allows for being conditional on covariates
#'
#' @param Geno A matrix of orginal genotype (# of individuals x # of SNPs)
#' @param withCovar A boolean variable indicating with or without covariates; the default is without covariates.
#' @param covar A matrix of covariates (# of individuals x # of covariates); the default is NULL
#'
#' @return A matrix of centralized genotype (# of individuals x # of SNPs)
#' 
genoCentralize = function(Geno,withCovar=FALSE,covar=NULL){
  N = dim(Geno)[1]
  if(withCovar){
    # Add intercept to the covariate matrix (one row of 1's)
    covar = cbind(rep(1, N),covar)
    # Calculate the H matrix of covariates
    covariateH = covar%*%solve(t(covar)%*%covar)%*%t(covar)
    Geno.scaled = scale((diag(N) - covariateH)%*%Geno)/sqrt(N-1)
  }
  if(!withCovar){
    Geno.scaled = scale(Geno)
  }
  return(Geno.scaled)
}


