####
# This file contains functions for calculating the optimal weights for
# summation-based p-value combination test statistics.
# The p-values are Z-scores, from which the core implementation is rooted.
# Creator: Hong Zhang 2023-04-18
# Last editions: ZWu
#  2023-08-27(HZ) (add normal approximated weights, re-write earlier version as sparse approximation)
#  2023-06-04 (add option is.posi.wts to function Optimal_Weights)
#  2023-04-21 (minor editorial revisions on headers/comments)
####

##################################
#Core function regression models #
##################################
#' Calculate the optimal weights of T = sum_i^n w_iT_i under regression models
#' @param g transformation function of the Z-scores
#' @param G genotype matrix (Nxn), N subjects, n genetic variants.
#' @param B vector of genetic effect sizes of causal variants (including all
#'          SNPs, which are causal when realized to be causal).
#' @param PI vector of estimated probability of each variant being signal.
#' @param trait indicator of "binary" (logistic regression) or "continuous"
#'        (linear regression).
#' @param nullmod the null model object obtained from the get_H function,
#'                which only needs to run once
#' @param X covariate matrix (Nxk), N subjects, k covariates. k>=1.
#'          If no covariates, X is an Nx1 matrix of 1s.
#' @param Y phenotype vector (Nx1). Note that Y~G will NOT be used to obtain 
#'          optimal weights.
#' @param is.posi.wts logical, if TRUE, the optimal weights are forced to be positive.
#' 
#' @return vectors of optimal weights w.
#'         wts_BE_sparse, wts_APE_sparse are the optimal weights based on sparse approximation, according to Bahardur efficiency and asymptotic power efficiency, respectively. 
#'         wts_BE_N, wts_APE_N are the optimal weights based on normal approximation
#' @details Note that he correlation matrix M is obtained under H0. It could be  different under H1 (which could be explored in future study)
Optimal_Weights_WIP <- function(g, G, B, PI, trait="binary", nullmod=NULL, X=NULL, Y=NULL, 
                    is.posi.wts=TRUE){
  if(is.null(nullmod)){
    nullmod = get_H(X, Y, trait)
  }
  
  if(trait=="binary"){
    Y0 = nullmod$Y0
    print("the dimension of Y0 is")
    print(length(Y0))
    #Whalf = diag(sqrt(Y0*(1-Y0)))
    #Gtilde = Whalf%*%G
    Gtilde = t(t(sqrt(Y0*(1-Y0)) * G))
    print("the dimensions of G tilde")
    print(dim(Gtilde))
    # GHG = t(Gtilde)%*%nullmod$I_H%*%Gtilde
    GHhalf = t(Gtilde)%*%nullmod$Hhalf
    print("the dimensino of nullmod hhalf is")
    print(dim(nullmod$Hhalf))
    print("the dimension of gHalf is")
    print(dim(GHhalf))
    GHG = t(Gtilde)%*%Gtilde - GHhalf%*%t(GHhalf)
    print("the dimension of GHG is")
    print(dim(GHG))
    #should be square matrix of 1,0
    # MU = GHG%*%B/(sqrt(diag(GHG)))
    Bstar = sqrt(diag(GHG))*B # standardized effect size
    #vector;
    print("the dimension of Bstar is")
    print(length(Bstar))
  }else if(trait=="continuous"){
    GHhalf = t(G)%*%nullmod$Hhalf
    # GHG = t(G)%*%nullmod$I_H%*%G
    GHG = t(G)%*%G - GHhalf%*%t(GHhalf)
    # MU = GHG%*%B/(nullmod$s0*sqrt(diag(GHG)))
    Bstar = sqrt(diag(GHG))*B/nullmod$s0 # standardized effect size
  }
  M = cov2cor(GHG)
  MU = as.matrix(Bstar*PI) # mean of theta
  MMU = M%*%MU # mean of marginal scores
  V = c(Bstar^2*PI*(1-PI))
  # faster way to compute M'VM, when V is a diagonal matrix
  MVM = t((M*V))%*%M
  
  r_tilde = get_r_tilde(g, MU=MMU, SD=diag(M+MVM))
  Sigma_APE_tilde = CovM_gXgY(g, MU1=MMU, MU2=MMU, M=M+MVM)
  Sigma_BE_tilde = CovM_gXgY(g, MU1=rep(0,length(MMU)), 
                             MU2=rep(0,length(MMU)), M=M)
  wts_BE_N = get_wts(Sigma_BE_tilde, r_tilde, is.posi.wts)
  wts_APE_N = get_wts(Sigma_APE_tilde, r_tilde, is.posi.wts)
  
  r = get_r(g, Bstar, PI)
  Sigma_BE = get_Sigma(g, Bstar, PI, M, hypo="H0")
  Sigma_APE = get_Sigma(g, Bstar, PI, M, hypo="H1")

  wts_BE_sparse = get_wts(Sigma_BE, r, is.posi.wts)
  wts_APE_sparse = get_wts(Sigma_APE, r, is.posi.wts)
  
  return(list(wts_BE_N=wts_BE_N$w_normalized, 
              wts_APE_N=wts_APE_N$w_normalized,
              wts_BE_sparse=wts_BE_sparse$w_normalized, 
              wts_APE_sparse=wts_APE_sparse$w_normalized))
}

####
# This file contains functions for calculating the optimal weights for
# summation-based p-value combination test statistics.
# The p-values are Z-scores, from which the core implementation is rooted.
# Creator: Hong Zhang 2023-04-18
# Last editions: ZWu
#  2023-06-04 (add option is.posi.wts to function Optimal_Weights)
#  2023-04-21 (minor editorial revisions on headers/comments)
####

##################################
#Core function regression models #
##################################
#' Calculate the optimal weights of T = sum_i^n w_iT_i under regression models
#' @param g transformation function of the Z-scores
#' @param G genotype matrix (Nxn), N subjects, n genetic variants.
#' @param B vector of genetic effect sizes of causal variants (including all
#'          SNPs, which are causal when realized to be causal).
#' @param PI vector of estimated probability of each variant being signal.
#' @param trait indicator of "binary" (logistic regression) or "continuous"
#'        (linear regression).
#' @param nullmod the null model object obtained from the get_H function,
#'                which only needs to run once
#' @param X covariate matrix (Nxk), N subjects, k covariates. k>=1.
#'          If no covariates, X is an Nx1 matrix of 1s.
#' @param Y phenotype vector (Nx1). Note that Y~G will NOT be used to obtain 
#'          optimal weights.
#' @param is.posi.wts logical, if TRUE, the optimal weights are forced to be positive.
#' 
#' @return vectors of optimal weights w.
#'         wts_BE, wts_APE are the optimal weights according to Bahardur efficiency and asymptotic power efficiency, respectively. 
#' @details Note that he correlation matrix M is obtained under H0. It could be  different under H1 (which could be explored in future study)
Optimal_Weights <- function(g, G, B, PI, trait="binary", nullmod=NULL, X=NULL, Y=NULL, 
                            is.posi.wts=TRUE){
  if(is.null(nullmod)){
    nullmod = get_H(X, Y, trait)
  }
  
  if(trait=="binary"){
    Y0 = nullmod$Y0
    #Whalf = diag(sqrt(Y0*(1-Y0)))
    #Gtilde = Whalf%*%G
    Gtilde = t(t(sqrt(Y0*(1-Y0)) * G))
    # GHG = t(Gtilde)%*%nullmod$I_H%*%Gtilde
    GHhalf = t(Gtilde)%*%nullmod$Hhalf
    GHG = t(Gtilde)%*%Gtilde - GHhalf%*%t(GHhalf)
    MU = GHG%*%B/(sqrt(diag(GHG)))
  }else if(trait=="continuous"){
    GHhalf = t(G)%*%nullmod$Hhalf
    # GHG = t(G)%*%nullmod$I_H%*%G
    GHG = t(G)%*%G - GHhalf%*%t(GHhalf)
    MU = GHG%*%B/(nullmod$s0*sqrt(diag(GHG)))
  }
  M = cov2cor(GHG)
  r = get_r(g, MU, PI)
  Sigma_BE = get_Sigma(g, MU, PI, M, hypo="H0")
  Sigma_APE = get_Sigma(g, MU, PI, M, hypo="H1")
  
  wts_BE = get_wts(Sigma_BE, r, is.posi.wts)
  wts_APE = get_wts(Sigma_APE, r, is.posi.wts)
  
  return(list(wts_BE_original=wts_BE$w, wts_BE=wts_BE$w_normalized, 
              wts_APE_original=wts_APE$w, wts_APE=wts_APE$w_normalized))
}

#' Calculate the null model Y~X outputs.
#' @param X covariate matrix (Nxk), N subjects, k covariates. k>=1, that is if no covariates, X is an Nx1 matrix of 1s.
#' @param Y phenotype vector (Nx1). Note that Y~G will NOT be used to obtain optimal weights.
#' @param trait indicator of "binary" (logistic regression) or "continuous" (linear regression).
#' @return H^1/2, where H is the projection matrix of a regression model. 
#'         Also outputs Y0, the null model predicted values for logistic regression 
#'         or s0 the null model standard error for linear regression. 
get_H = function(X, Y, trait="binary"){
  # N = dim(X)[1]
  if(trait=="binary"){
    mod0 = glm(Y~X, family="binomial")
    Y0 = mod0$fitted.values
    #Whalf = diag(sqrt(Y0*(1-Y0)))
    #Xtilde = Whalf%*%X
    Xtilde = t(t(sqrt(Y0*(1-Y0)) * X))
    # Htilde = Xtilde%*%solve(t(Xtilde)%*%Xtilde)%*%t(Xtilde)
    Hhalf = Xtilde%*%t(chol(solve(t(Xtilde)%*%Xtilde)))
    res = Y - Y0
    return(list(Hhalf=Hhalf, Y0=Y0, res=res))
  }else if(trait=="continuous"){
    mod0 = glm(Y~X, family="gaussian")
    Y0 = mod0$fitted.values
    s0 = sd(mod0$residuals)
    Hhalf = X%*%t(chol(solve(t(X)%*%X)))
    # H = X%*%solve(t(X)%*%X)%*%t(X)
    # GHG = t(G)%*%(diag(1,N,N)-H)%*%G
    res = Y - Y0
    return(list(Hhalf=Hhalf, s0=s0, res=res))
  }
}
#######################################
#Core functions Gaussian mixture model#
#######################################

#' Calculate the optimal weights of T = sum_i^n w_iT_i under Gaussian mixture model
#' @param Sigma covariance matrix of T_1,...,T_n.
#' @param r vector of E_1(T_i) - E_0(T_i), i=1,...,n.
#' @param is.posi.wts logical, if TRUE, the optimal weights are forced to be positive.
#' @return vector of optimal weights w. 
#'         
#' @details The weights are normalized to have mean absolute values being 1.
#'          Or, the sum of absolute values being n. 
#' @example     
#' Sigma = solve(matrix(c(1,-0.5,-0.5,1),2,2))
#' r = c(1,0)
#' get_wts(Sigma, r); #force weights to be positive by default
#' get_wts(Sigma, r, is.posi.wts=FALSE); #negative weights ar possible
#' 
get_wts = function(Sigma, r, is.posi.wts=TRUE){
  w = solve(Sigma)%*%r
  if(is.posi.wts){
    w = w*(w>0)
  }
  return(list(w=w, w_normalized=w/mean(abs(w)))); 
}

#' Calculate the vector of E_1(T_i) - E_0(T_i), i=1,...,n, 
#' where T_i=g(Z_i), Z_i = Z_0i + mu_i*C_i, C_i ~ Bern(pi_i), Z_0i~N(0,1)
#' @param g transformation function of Z 
#' @param MU vector of mean mu_i, i=1,...,n.
#' @param PI vector of probability pi_i, i=1,...,n.
#' @return vector of E_1(T_i) - E_0(T_i).
get_r = function(g, MU, PI){
  n = length(MU)
  sapply(1:n, function(x)E_T_mix(g, MU[x], PI[x])) - E_T_mix(g, 0, 0)
}

#' Calculate the vector of E_1(T_i) - E_0(T_i), i=1,...,n, 
#' where T_i=g(Z_i), Z_i = Z_0i + Ztilde_i, Ztilde_i ~ N(mu_i,sd_i), Z_0i~N(0,1)
#' @param g transformation function of Z 
#' @param MU vector of mean mu_i, i=1,...,n.
#' @param SD vector of standard deviation sd_i, i=1,...,n.
#' @return vector of E_1(T_i) - E_0(T_i).
get_r_tilde = function(g, MU, SD){
  n = length(MU)
  sapply(1:n, function(x)E_gX_p(g, mu=MU[x], p=1, sigma=SD[x])) - E_gX_p(g, mu=0, p=1, sigma=1)
}

#' Calculate the covariance matrix of T_1,...,T_n, 
#' where T_i=g(Z_i), Z_i = Z_0i + mu_i*C_i, C_i ~ Bern(pi_i),
#' Z_0 ~ MVN(0, M), M is a positive-definite correlation matrix
#' @param g transformation function of Z 
#' @param MU vector of mean mu_i, i=1,...,n.
#' @param PI vector of probability pi_i, i=1,...,n.
#' @param M correlation matrix of Z_0
#' @param hypo indicator "H0" means null hypothesis, MU=0, PI=0; "H1" means alternative
#' @return covariance matrix of T_1,...,T_n
get_Sigma = function(g, MU, PI, M, hypo="H1"){
  n = length(MU)
  if(hypo=="H1"){
    CovMT_mix(g, MU, PI, M)
  }else{
    CovMT_mix(g, rep(0,n), rep(0,n), M)
  }
}


##################################################
# Helpers deal with Gaussian mixture distribution#
##################################################

# Calculate E[T_i], where T_i=g(Z_i), 
# Z_i = Z_0i + mu_i*C_i, C_i ~ Bern(pi_i)
E_T_mix = function(g, mu, pi){
  pi*E_gX_p(g,mu,1) + (1-pi)*E_gX_p(g,0,1)
}

# Calculate Var[T_i], where T_i=g(Z_i), 
# Z_i = Z_0i + mu_i*C_i, C_i ~ Bern(pi_i)
Var_T_mix = function(g, mu, pi){
  pi*E_gX_p(g, mu, 2) + (1-pi)*E_gX_p(g, 0, 2)  - (E_T_mix(g, mu, pi))^2
}

# Calculate the covariance matrix Cov[T_i, T_j], 
# where T_i=g(Z_i), T_j=g(Z_j), 
# Z_i = Z_0i + mu_i*C_i, C_i ~ Bern(pi_i), 
# Z_j = Z_0j + mu_j*C_j, C_j ~ Bern(pi_j),
# Cov[Z_0i, Z_0i] = M_ij
# MU and PI are the vectors mu and pi.
CovMT_mix = function(g, MU, PI, M, ORD=1:8){
  n = length(MU)
  PI_M = PI%*%t(PI)
  PI_M1 = PI%*%t(1-PI)
  PI_M2 = t(PI_M1)
  PI_M3 = (1-PI)%*%t(1-PI)
  Cov_M = CovM_gXgY(g, MU, MU, M, ORD=1:8)
  Cov_M1 = CovM_gXgY(g, MU, rep(0,n), M, ORD=1:8)
  Cov_M2 = t(Cov_M1)
  Cov_M3 = CovM_gXgY(g, rep(0,n), rep(0,n), M, ORD=1:8)
  M_out = PI_M*Cov_M + PI_M1*Cov_M1 + PI_M2*Cov_M2 + PI_M3*Cov_M3
  diag(M_out) = sapply(1:n, function(x)Var_T_mix(g, MU[x], PI[x]))
  M_out
}


########################################
# Helpers deal with Normal distribution#
########################################

# Calculate E[g(X)^p] where X~N(mu,sigma)
E_gX_p = function(g, mu, p, sigma=1){
  integrate(function(x)g(x)^p*dnorm((x-mu)/sigma),mu-8*sigma,mu+8*sigma)$value/sigma
}


# Hermite polynomials order (ord) = 1 -- 8, with shift mu
hermite = function(x, mu, ord){
  if(ord==1){
    x-mu
  }else if(ord==2){
    (x-mu)^2 - 1
  }else if(ord==3){
    (x-mu)^3-3*(x-mu)
  }else if(ord==4){
    (x-mu)^4-6*(x-mu)^2+3
  }else if(ord==5){
    (x-mu)^5 - 10*(x-mu)^3 + 15*(x-mu)
  }else if(ord==6){
    (x-mu)^6-15*(x-mu)^4+45*(x-mu)^2-15
  }else if(ord==7){
    (x-mu)^7 - 21*(x-mu)^5 + 105*(x-mu)^3 - 105*(x-mu)
  }else if(ord==8){
    (x-mu)^8-28*(x-mu)^6+210*(x-mu)^4-420*(x-mu)^2+105
  }
}



# Calculate Covariance matrix whose elements are Cov[g(Xi),g(Yj)]
# where X = Z + MU1 ~ MVN(MU1, M), Y = Z + MU2 ~ MVN(MU2, M)
# M is a positive-definite correlation matrix
CovM_gXgY = function(g, MU1, MU2, M, ORD=1:8){
  coef1 = sapply(MU1, function(z)sapply(ORD, function(y)integrate(function(x)g(x)*dnorm(x-z)*hermite(x,z,y), z-8,z+8)$value))
  coef2 = sapply(MU2, function(z)sapply(ORD, function(y)integrate(function(x)g(x)*dnorm(x-z)*hermite(x,z,y), z-8,z+8)$value))
  M_out = matrix(0, ncol=length(MU1), nrow=length(MU1))
  for(i in 1:length(ORD)){
    ord = ORD[i]
    M_out = M_out + as.matrix(coef1[ord,])%*%t(as.matrix(coef2[ord,]))*M^ord/factorial(ord)
  }
  M_out
}


#########################################
# Helpers deal transformation function g#
#########################################

g_GFisher_two = function(x, df){
  qchisq(log(2)+pnorm(abs(x),lower.tail = FALSE,log.p=TRUE), 
         df = df, lower.tail = F,log.p=TRUE)
}

g_GFisher_one = function(x, df){
  qchisq(pnorm(x,lower.tail = FALSE,log.p=TRUE), 
         df = df, lower.tail = F,log.p=TRUE)
}

g_Burden = function(x) x


#================== 
#  Below are useful functions for calculating inidividual components of matrices
#   they are not directly called by the main functions b/c of more effeciient implementation by linear algibra calculations
#  (tested and confirmed with the above implementation)
#==================


# Calculate Var[g(X)] where X~N(mu,1)
Var_gX = function(g, mu){
  E_gX_p(g, mu, 2) - (E_gX_p(g, mu, 1))^2
}

# Calculate Covariance matrix of [g(X1),...,g(Xn)]
# where X ~ MVN(MU, M), M is a positive-definite 
# correlation matrix
CovM_gX = function(g, MU, M, ORD=1:8){
  coef = sapply(MU, function(z)sapply(ORD, function(y)integrate(function(x)g(x)*dnorm(x-z)*hermite(x,z,y), z-8,z+8)$value))
  M_out = matrix(0, ncol=length(MU), nrow=length(MU))
  for(i in 1:length(ORD)){
    ord = ORD[i]
    M_out = M_out + as.matrix(coef[ord, ])%*%t(as.matrix(coef[ord, ]))*M^ord/factorial(ord)
  }
  M_out
}

# Calculate Cov[T_i, T_j], where T_i=g(Z_i), T_j=g(Z_j), 
# Z_i = Z_0i + mu_i*C_i, C_i ~ Bern(pi_i), 
# Z_j = Z_0j + mu_j*C_j, C_j ~ Bern(pi_j),
# Cov[Z_0i, Z_0j] = M_ij (=rho in the function input options)         
# Note that Cov_TiTj_mix cannot reduce to Var_T_mix by simply 
# letting rho=1 because C_i and C_j are independent random variables.
Cov_TiTj_mix = function(g, mu1, pi1, mu2, pi2, rho, ORD=1:8){
  pi1*pi2*Cov_gXigXj(g,mu1,mu2,rho,ORD) + 
    pi1*(1-pi2)*Cov_gXigXj(g,mu1,0,rho,ORD) +
    pi2*(1-pi1)*Cov_gXigXj(g,0,mu2,rho,ORD) +
    (1-pi1)*(1-pi2)*Cov_gXigXj(g,0,0,rho,ORD)
}


# Calculate Cov[g(Xi), g(Xj)] where 
# Xi~N(mu1,1), Xj~N(mu2,1), cor(Xi,Xj)=rho
Cov_gXigXj = function(g, mu1, mu2, rho, ORD=1:4){
  coef_1 = sapply(ORD, function(y)integrate(function(x)g(x)*dnorm(x-mu1)*hermite(x,mu1,y), mu1-8,mu1+8)$value)
  coef_2 = sapply(ORD, function(y)integrate(function(x)g(x)*dnorm(x-mu2)*hermite(x,mu2,y), mu2-8,mu2+8)$value)
  sum(coef_1*coef_2*rho^ORD/factorial(ORD))
}

# #############################################################
# 
# getdelta = function(b2, sigma){
#   return(b2/sigma^2)
# }
# 
# trapezoid = function(x, y){
#   y1 = y[1:(length(y)-1)]
#   y2 = y[2:length(y)]
#   y_trap = (y1+y2)/2
#   sum(y_trap*diff(x))
# }
# 
# normalizeGeno = function(x){
#   xmax = max(x)
#   xmin = min(x)
#   2 - 2*(xmax-x)/(xmax-xmin)
# }
