###
# This file contains experimental code snippets for implementing the GFisher test in Hail
# Strategy is to follow the SKAT test in Hail for data processing
# and then implement calculations for Zscore, LD matrix, and GFisher test
# Creator: ZWu 2024-04-10
# Modified: TYakovenko 2024-04-22

#this file contains calculations done using Hail for Zscore, Pvalues, GFisher statistics, and M matrix using Hail
#now use those calculations and insert the into the old pipeline which utilizes numpy
###

import hail as hl
import p_GFisherNumpy as pgfnp
import time
import exampleHt as eht

start = time.time()
hl.version()
hl.init(log="hailDistributedProcessing.log")

#using the example given in the Hail source code for Skat, output the ht matrix table once it has been processed
#eht.saveHt() #this function only needs to be run once to save everything under ht.mt

ht = hl.read_table("hailMatrices/ht.mt")

#function to return a hail matrix with all the necessary pGFisher inputs
def findGFM (ht):
    GHG = ht.G.T @ ht.G - ht.G.T @ ht.covmat_Q @ ht.covmat_Q.T @ ht.G  # GHG is the covariance matrix for G given X (covariates)
    scores = (ht.y_residual @ ht.G) / hl.sqrt(ht.s2)  # The marginal score statistics
    sd_GHG = hl.sqrt(hl.nd.diagonal(GHG))
    Zscores = scores / sd_GHG
    pvalues = Zscores.map(lambda x: 2 * hl.pnorm(-hl.abs(x), lower_tail=True))
    sd_GHG_m = sd_GHG.reshape(
        (1, sd_GHG.shape[0]))  # method reshape doesn't allign with the calss Float here(?????????????????????)
    # sd_GHG_m.show()
    outer_sd_GHG = sd_GHG_m.T @ sd_GHG_m  # outer product of the diagonal elements
    #^requires matrices and reshaping
    # outer_sd_GHG.show()
    M_s = GHG / outer_sd_GHG
    #print("The M matrix is")
    #M_s.show()  # M_s is the correlation matrix for G given X (covariates)
    #annotate the global fields in the input data -> there is NO return here; we are adding onto a matrix in place
    #no annotation in place but we can replace it such that there are no returns
    # Annotate multiple global fields to the MatrixTable
    def g_quadr(x):  # quadradic function
        return x ** 2

    g = g_quadr
    gf_quadr = (Zscores.map(lambda x: g(x)) * ht.weight).sum(0)  # directly use weights in ht
    #gf_quadr.show()

    # We can get SKAT by using Zscores, gquadr, and proper weights being sd_GHG.map(lambda x: x**2)
    (scores.map(lambda x: g(x)) * ht.weight).sum(0)  # This gives the SKAT statistic, same as above
    #(Zscores.map(lambda x: g(x)) * sd_GHG.map(lambda x: x ** 2)).sum(0).show()  # This gives the SKAT statistic, same as above

    ##By identity function, we get Liptak test statistics
    # This corresponds to the burden test as a special case
    def g_ident(x):  # identity function
        return x

    g = g_ident
    gf_ident = (Zscores.map(lambda x: g(x)) * ht.weight).sum(0)  # directly use weights in ht
    #gf_ident.show()

    ##By GFisher transformation function, we get GFisher test statistics
    def g_GFisher_two(x, df=2):
        return hl.qchisqtail(hl.log(2) + hl.pnorm(-hl.abs(x), log_p=True), df=df, lower_tail=False, log_p=True)

    # Fisher's combination test statistic
    g = g_GFisher_two
    gf_GFisher = (Zscores.map(lambda x: g(x)) * ht.weight).sum(0)  # directly use weights in ht
    return pvalues, gf_GFisher, M_s


#the below were calculated using functions above developed by Wu,Z.
pvals1, gf1, M1 = findGFM(ht) #M_s is directly called in p_GFisher but pvals need to be recalibrated

#now insert the values into an existing pipeline to get the individual SNP values
#convert the values into nupmy to accomodate the old pipeline approach
#TODO check that the calulcations work correct with gf being an array for each gene instead of being a single value
pvals = pvals1.collect()
gf = gf1.collect() #TODO: check how the aggregation function works and see if it is compatible with dist computing
M = M1.collect()
'''
note that p_GFisher function which is implemented in p_GFisherNumpy requires inputs for ONE gene only, if we want to get the results for two of the genes, we need to run an iterative loop
this approach is not idea computationally but will overall work
'''
def calSNPPVals (gf, M, df=1, w=1):
    '''
    A function to handle calculating multiple p values for the genes since the original implementation of P_Gfisher requires input only for one gene
    Double-check the default parameters below to meet the particular needs;
    These p values are calculated using the MR method
    :param gf: gFisher statistic calculated using the stat function or Hail approach
    gf contains stats for each gene
    :param M: correlation matrix
    M contains two matricies
    Notice that the length of the list M and list of gf must match and represent the number of genes to be evaluated
    See conversion of M and gf above from the ht calculations performed using Hail
    :return: list of P values corresponding to each gene
    '''
    allPVals = []
    for i in range (0, len(M)):
        pVal = pgfnp.p_GFisher(gf[i], df, w, M=M[i], method="MR", nsim=5000)  # pvalue using p-GFisher
        allPVals.append(pVal)
    return allPVals

res = calSNPPVals(gf, M)
print(res)
end = time.time()
print(end-start)
hl.stop() #stop the hail session so that there are on conflicts with other files
#TODO include all the required input types for various functions and comment on how to do that changes if necessary
#TODO: double check that the calculations are correct

'''
Overall note: we do not use any numpy functionality for conversion here directly but the code 
will still not scale due to the fact that the core calculations are not scallable
'''
