'''
This file contains the Python functions for the GFisher test, converted based on R code
GFisher_v1.R
The file uses numpy solution which is now outdates as it is not scallable
The solution below still works on smaller datasets and files
Update:
    2024-02-17: Created by Taya Yakovenko
'''

import numpy as np
#from hail import pnorm #TODO see if this needs to be used over the other way to find pnorm
import hail as hl
from numpy import linalg as la
from scipy.integrate import quad
from scipy.linalg import eigvalsh, eig
from scipy.sparse import diags
from scipy.stats import norm, chi2, gamma #TODO provide alternative libraries to use for calculations and describe the difference between implementation
from sklearn.utils.extmath import safe_sparse_dot
import npd


def stat_GFisher(p, df=2, w=1):
    '''

    Compute the GFisher test statistics.

    :param p: A vector of input p-values of the GFisher test.
    :param df: Degrees of freedom for inverse chi-square transformation for each p-value.
               It can be a vector of the same length as p, indicating each transformation function
               might have a different df, or a single number, indicating the same degrees of freedom for all.
    :param w: Vector of non-negative weights for each p-value. Default is 1 (equal weights).
    :return: The GFisher test statistics.
    :rtype: float
    :details: The statistic is calculated based on normalized weights, i.e., w/mean(w).
    :author: Hong Zhang
    :examples:
        n = 10
        pval = np.random.rand(n)
        fisher_stat(pval, df=2, w=1)
        fisher_stat(pval, df=[2]*n, w=[1]*n)
        fisher_stat(pval, df=list(range(1, n+1)), w=list(range(1, n+1)))
    :export:
         # Example 1:
    p_values = [0.1, 0.05, 0.01]
    fisher_statistic = stat_GFisher(p_values)
    print("Fisher Statistic (Default arguments):", fisher_statistic)

    # Example 2:
    # Providing custom degrees of freedom and weights
    custom_df = [3, 3, 3]
    custom_weights = [0.2, 0.3, 0.5]
    fisher_statistic_custom = stat_GFisher(p_values, df=custom_df, w=custom_weights)
    print("Fisher Statistic (Custom arguments):", fisher_statistic_custom)

    # Generate random p-values
    n = 10
    pval = np.random.rand(n)

    # Example 1: Using default degrees of freedom (df=2) and weights (w=1)
    fisher_statistic_default = stat_GFisher(pval)
    print("Fisher Statistic (Default arguments):", fisher_statistic_default)

    # Example 2: Providing degrees of freedom as rep(2, n) and weights as rep(1, n)
    df_rep = np.repeat(2, n)
    w_rep = np.repeat(1, n)
    fisher_statistic_rep = stat_GFisher(pval, df=df_rep, w=w_rep)
    print("Fisher Statistic (Custom df and w):", fisher_statistic_rep)

    # Example 3: Providing degrees of freedom as 1:n and weights as 1:n
    df_n = np.arange(1, n+1)
    w_n = np.arange(1, n+1)
    fisher_statistic_n = stat_GFisher(pval, df=df_n, w=w_n)
    print("Fisher Statistic (Custom df and w):", fisher_statistic_n)

    '''
    if isinstance(w, (list, np.ndarray)):
        w = len(p) * w / np.sum(w)
    else:
        w = 1
    if np.all(df == 2):
        pp_trans = -2 * np.log(p)
    else:
        pp_trans = chi2.isf(p, df=df)
        na_id = np.where(np.isnan(pp_trans))[0]
        if len(na_id) > 0:
            pp_trans[na_id] = chi2.ppf(1 - p[na_id], df=df[na_id])
    fisher_stat = np.sum(w * pp_trans)
    return fisher_stat



def p_GFisher(q, df, w, M, p_type="two", method="HYB", nsim=None, seed=None):
  """
  Calculates p-value using Fisher's exact test with various methods.
  Depends: ....

  Args:
      q: Test statistic (float).
      df: Degrees of freedom (array-like of ints).
      w: Weights for observations (array-like of floats).
      M: correlation matrix of the Zscores from which the input p-values were obtained.
      p_type: Tail of the test ("two" for two-sided, other for one-sided).
      method: Method for calculating p-value ("MR" for Monte Carlo, "HYB" for hybrid, "Brown" for Brown's method).
      nsim: Number of simulations for Monte Carlo method (int, default=None).
      seed: Random seed for Monte Carlo simulation (int, default=None).

  Returns:
      The calculated p-value (float).

    Examples:
    np.random.seed(123)
    # Define the dimensions
    n = 10
    # Create the matrix M
    M = 0.3 * np.ones((n, n)) + np.diag(0.7 * np.ones(n))
    # Generate zscore -> zscore; pval, gf1 will not always match the r code due to random generation
    zscore = np.random.normal(size=n) @ np.linalg.cholesky(M)
    # Calculate p-values
    pval = 2 * (1 - norm.cdf(np.abs(zscore)))

    gf1 = stat_GFisher(pval, df=2, w=1)
    gf2 = stat_GFisher(pval, df=np.arange(1, n + 1), w=np.arange(1, n + 1))

    result1 = p_GFisher(gf1, df=2, w=1, M=M, method="HYB")
    result2 = p_GFisher(gf1, df=2, w=1, M=M, method="MR", nsim=5e4)
    result3 = p_GFisher(gf2, df=list(range(1, n + 1)), w=list(range(1, n + 1)), M=M, method="HYB")
    result4 = p_GFisher(gf2, df=11, w=11, M=M, method="MR")


    print("Result 1 (HYB method, df=2, w=1):", result1)
    print("Result 2 (MR method, df=2, w=1, nsim=5e4):", result2)
    print("Result 3 (HYB method, df=1:n, w=1:n):", result3)
    print("Result 4 (MR method, df=2, w=1, nsim=5e4):", result4)
  """
  n = M.shape[1]  # Number of columns in contingency table

  #assure that the function runs with arrays and constants; constants will be turned into arrays of repeating values of the given constant
  if isinstance(df, int) or isinstance(df, float):
      df = np.full(n, df)
  if isinstance(w, int) or isinstance(w, float):
      w = np.full(n, w)

  # Normalize weights
  w = np.multiply(n,w) / np.sum(w)

    #MR method implementation
  if method == "MR":
    #raise NotImplementedError("Monte Carlo simulation method (MR) not implemented.")
    nsim = int(nsim)
    #check if M is positive definite and if not change it as needed
    if(not npd.NPD.isPD(M)):
        M = npd.NPD.nearestPD(M)
    #now working with nearest positive definite as needed
    try:
        M_chol = np.linalg.cholesky(M)
    except np.linalg.LinAlgError:
        M = np.matrix(np.linalg.matrix_rank(M, tol=None))  # Convert to full rank matrix
        M_chol = np.linalg.cholesky(M)
    if isinstance(seed, (int, float)):
        np.random.seed(seed)
    znull = np.random.normal(size=(nsim, n)) @ M_chol
    if p_type == "two":
        pnull = 2 * norm.cdf(-np.abs(znull))
    else:
        pnull = norm.cdf(znull, loc=0, scale=1, lower=False)
    if np.std(df) == 0:
        if np.all(df == 2):
            pp_trans = -2 * np.log(pnull)
        else:
            pp_trans = chi2.ppf(1 - pnull, df[0])
    else:
        pp_trans = np.array([chi2.ppf(1 - x, df[i]) for i, x in enumerate(pnull)])
    fishernull = np.sum(pp_trans * w, axis=1)
    MM = np.array([np.mean(fishernull ** x) for x in range(1, 5)])
    mu = MM[0]
    sigma2 = (MM[1] - MM[0] ** 2) * nsim / (nsim - 1)
    cmu3 = MM[2] - 3 * MM[1] * MM[0] + 2 * MM[0] ** 3
    cmu4 = MM[3] - 4 * MM[2] * MM[0] + 6 * MM[1] * MM[0] ** 2 - 3 * MM[0] ** 4
    gm = cmu3 / sigma2 ** (3 / 2)
    kp = cmu4 / sigma2 ** 2 + 3
    a = 9 * gm ** 2 / (kp - 3) ** 2
    pval = gamma.sf((q - mu) / np.sqrt(sigma2) * np.sqrt(a) + a, a=a, scale=1)

    #Browns and HYB methods
  else:
    # Calculate GM statistic based on chosen method,
    GM = getGFisherGM(df, w, M, p_type)

    # Calculate mean and variance
    mu = np.sum(w * df)
    sigma2 = np.sum(GM)

    if method == "HYB":
      # Hybrid method (requires additional function getGFisherlam)
      out_GM = getGFisherlam(df, w, M, GM, p_type)
      lam = out_GM["lam"]
      c2 = np.sum(lam**2)
      c3 = np.sum(lam**3)
      c4 = np.sum(lam**4)

      # Calculate skewness and kurtosis
      gm = np.sqrt(8) * c3 / c2**(3/2)
      kp = 12 * c4 / c2**2 + 3
    else:
      # Brown's method
      a = mu**2 / sigma2

    # Calculate p-value using pgamma function
    a = 9 * gm**2 / ((kp - 3)**2)
    pval = 1 - norm.cdf(q, loc=mu, scale=np.sqrt(abs(sigma2)) * np.sqrt(a))  # One-tailed test
    if p_type == "two":
      pval = 2 * min(pval, 1 - pval)  # Two-tailed test

  return pval

#####################################################
#helper functions
#####################################################

def getGFisherlam(D, w, M, GM, p_type = "two"):
    '''
    Calculate the lambda
    :param D: degrees of freedom as a vector
    :param w: weights, same dimensions as degrees of freedom but a square matrix
    :param M: correlation matrix of the Zscores from which the input p-values were obtained.
    :param GM: Fished information matrix calcualted from getGfisherGM
    :param p_type: calculation type
    :return: lambda calculation

    Examples:
    D = [2,2,2] # !!! In this particular function call degrees of freedom can NOT be a scalar.
    #However, when this function is called in p_GFisher the scalars get converted to appropritate arrays with needed dimensions
    w = np.array([[0.5, 0.5, 0.5], [0.5, 0.5, 0.5], [0.5, 0.5, 0.5]])
    M = np.array([[1, 0.5, 0.3], [0.5, 2, 0.1], [0.3, 0.1, 3]])  # M is a correlation matrix
    GM = np.eye(3)
    result = getGFisherlam(D, w, M, GM)
    print(result)

    '''
    # p_type is not used in function definintion in R but it was kept here for consistency with the origignal code
    n = M.shape[0]
    # redifine DM such that D is now a constant and use pmin calculations
    # check for pmin calc for DM
    # Calculate diagonal matrices of square root of D and absolute value of GM elements
    #D_sqrt = np.diag(np.sqrt(D))
    #abs_GM = np.abs(GM)

    # Calculate DM using element-wise minimum between row-wise and column-wise minimums of D
    # Calculate M_tilde using element-wise operations
    #M_tilde = np.multiply(np.multiply(D_sqrt, np.sqrt(0.5 * abs_GM / DM)), np.sign(M))

    DM = np.minimum(np.repeat(D, n).reshape(n, n), np.repeat(D, n).reshape(n, n).T)
    M_tilde = np.sqrt(np.abs(GM) / DM / 2) * np.sign(M)
    M_tilde = np.where(M_tilde > 1, 0.999, M_tilde)

    if np.any(np.linalg.eigvalsh(M_tilde) < 1e-10) and (not npd.NPD.isPD(M_tilde)):
        M_tilde = np.asmatrix(npd.NPD.nearestPD(M_tilde))

    M_tilde_chol = np.linalg.cholesky(M_tilde)
    # Changed!!! Commented w.reshape since it tarnsform w into vector that won't fly for all functions below, w should be a matrix
    #w = w.reshape(M_tilde.shape[0], 1)

    # Weighted Cholesky decomposition (WM_chol)
    # Changed!!! np.diag(np.sqrt(w)) creates a vector so @ operand return a vector too, failed below
    sqrtW = np.sqrt(w)[:, None] * np.ones((n, n))
    WM_chol = M_tilde_chol @ np.diag(np.diag(sqrtW))
    # tb = np.bincount(D)
    # tbval = np.nonzero(tb)[0]
    #lam = eigvalsh(safe_sparse_dot(WM_chol.T, WM_chol))
    lam, _ = eig(safe_sparse_dot(WM_chol.T, WM_chol))
    # Changed !!! Did not check but D argument cannot be scalar, must be a matrix (vector)
    if np.max(D) > 1:
        for i in range(2, np.max(D) + 1):
            arr = []
            for d in D:
                if d < 1:
                    arr.append(0)
                else:
                    arr.append(1)
            Ai = diags(arr, 0)
            lam = np.concatenate((lam, eigvalsh(safe_sparse_dot(safe_sparse_dot(WM_chol, Ai), WM_chol))))

    lam = lam[lam > 1e-10]
    return {"lam": lam}
def getGFisherGM(D, w, M, p_type="two"):
    '''
Calculate a Fisher information matrix (GM) in the context of network analysis.

  **Inputs:**

  * D: A list representing degrees of freedom for each node in the network.
  * w: A list of weights associated with the nodes in the network.
  * M: The network adjacency matrix.
  * p_type (optional): A string specifying the type of p-values used ("two" for two-sided, "one" for one-sided). Defaults to "two".

  **Output:**

  * The function returns the calculated Fisher information matrix (GM).

  **Example Usages:**

  ```
  D = [1, 2, 3]  # Values for degrees of freedom
  w = [0.5, 0.3, 0.2]  # Weights
  M = np.random.randn(3, 3)  # Example matrix M
  p_type = "two"  # Two-sided input p-values
  result = getGFisherGM(D, w, M, p_type)
  print(result)

    '''

    n = M.shape[0]

    def integrand(x, y, m, p_type):
        if p_type == "two":
            return qchisq(pchisq(x ** 2, df=1), df=y) * dnorm(x) * (m ** 2 / 2) * (x ** 2 - 1)
        else:
            #hail details -> pnorm returns an expresion instead of a value
            # TODO: figure out the pnorm here
            #pnormX = hl.eval(pnorm(x))
            pnormX = hl.eval(x)
            return qchisq(pnormX, df=y) * dnorm(x) * m * x

    def qchisq(p, df):
        return chi2.ppf(p, df)

    def pchisq(q, df):
        return chi2.cdf(q, df)

    def dnorm(x):
        return norm.pdf(x)

    #TODO: consider another approximation method to speed up the calculations
    if p_type == "two":
        coeff2 = np.array(
            [quad(lambda x: integrand(x, y, M[i, j], p_type), -8, 8)[0] for i, y in enumerate(D) for j in range(n)])
        coeff2 = coeff2.reshape((n, n))

        coeff4 = np.array(
            [quad(lambda x: integrand(x, y, M[i, j], p_type), -8, 8)[0] for i, y in enumerate(D) for j in range(n)])
        coeff4 = coeff4.reshape((n, n))

        coeff6 = np.array(
            [quad(lambda x: integrand(x, y, M[i, j], p_type), -8, 8)[0] for i, y in enumerate(D) for j in range(n)])
        coeff6 = coeff6.reshape((n, n))

        coeff8 = np.array(
            [quad(lambda x: integrand(x, y, M[i, j], p_type), -8, 8)[0] for i, y in enumerate(D) for j in range(n)])
        coeff8 = coeff8.reshape((n, n))

        GM = M.T ** 2 / 2 * coeff2 @ coeff2.T + M.T ** 4 / 24 * coeff4 @ coeff4.T + M.T ** 6 / 720 * coeff6 @ coeff6.T + M.T ** 8 / 40320 * coeff8 @ coeff8.T
        GM = np.diag(np.sqrt(np.multiply(D, 2)) * w) @ np.corrcoef(GM) @ np.diag(np.sqrt(np.multiply(D, 2)) * w)
    else:
        coeff3 = np.array(
            [quad(lambda x: integrand(x, y, M[i, j], p_type), -8, 8)[0] for i, y in enumerate(D) for j in range(n)])
        coeff3 = coeff3.reshape((n, n))

        coeff4 = np.array(
            [quad(lambda x: integrand(x, y, M[i, j], p_type), -8, 8)[0] for i, y in enumerate(D) for j in range(n)])
        coeff4 = coeff4.reshape((n, n))

        coeff2 = np.array(
            [quad(lambda x: integrand(x, y, M[i, j], p_type), -8, 8)[0] for i, y in enumerate(D) for j in range(n)])
        coeff2 = coeff2.reshape((n, n))

        coeff1 = np.array(
            [quad(lambda x: integrand(x, y, M[i, j], p_type), -8, 8)[0] for i, y in enumerate(D) for j in range(n)])
        coeff1 = coeff1.reshape((n, n))

        GM = M * coeff1 @ coeff1.T + M ** 2 / 2 * coeff2 @ coeff2.T + M ** 3 / 6 * coeff3 @ coeff3.T + M ** 4 / 24 * coeff4 @ coeff4.T
        GM = np.diag(np.sqrt(2 * D) * w) @ np.corrcoef(GM) @ np.diag(np.sqrt(2 * D) * w)

    return GM



